unit sDBComboBox;
(************************************************************************
sStyle components package 
Copyright (c) 2002 Serge Goncharov. All rights reserved.
                            Ukraine, Odessa
                            http://www.geocities.com/sstylish2002/
                            Serge@HyBce.com
 Version:	1.2, 24-02-2002
*************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, DBCtrls, sStyleUtil, StrUtils, sConst, sUtils, sGraphUtils;

type
  TsDBComboBox = class(TDBComboBox)
  private
    { Private declarations }
    FColorArrow: TColor;
    FColorButton: TColor;

    FButtonWidth: Integer;

    procedure SetColors (Index: Integer; Value: TColor);

    function GetButtonRect: TRect;
    procedure PaintButton;

    procedure RedrawBorders;

    procedure WMPaint (var Message: TWMPaint); message WM_PAINT;
    procedure CNCommand (var Message: TWMCommand); message CN_COMMAND;
    procedure MEnter(var Msg:TMessage); message CM_ENTER;
    procedure MExit(var Msg:TMessage); message CM_EXIT;
  protected
    { Protected declarations }
    FActive : boolean;
    FCharsInCode : integer;
    FSQL : TStrings;
    FsStyle : TsStyleDB;
    FUseItemIndex : boolean;
    BorderStyle : TFormBorderStyle;

    procedure SetSQL(Value: TStrings);
    procedure SetActive(Value : boolean);

    procedure WndProc (var Message: TMessage); override;
  public
    { Public declarations }
    FChildHandle: HWND;
    FDefListProc: Pointer;
    FListHandle: HWND;
    FListInstance: Pointer;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure Invalidate; override;
    function GetCurrentKod : integer;
    function GetCurrentName : string;
    function IndexOfKod(s : string) : integer;
    function Generate : integer; virtual;
    property Color;
    procedure InvalidateSelection;
  published
    { Published declarations }
    property ColorArrow: TColor index 0 read FColorArrow write SetColors default clBlack;
    property ColorButton: TColor index 1 read FColorButton write SetColors default $00C5D6D9;

    property UseItemIndex : boolean read FUseItemIndex write FUseItemIndex;
    property ButtonWidth : integer read FButtonWidth write FButtonWidth;
    property ItemIndex;

    property sStyle:TsStyleDB read FsStyle write FsStyle;
    property Active:boolean read FActive write SetActive;
    property SQL:TStrings read FSQL write SetSQL;
    property CharsInCode:integer read FCharsInCode write FCharsInCode;
  end;

implementation

constructor TsDBComboBox.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csFixedHeight, csFramed] + [csOpaque];
  TControlCanvas(Canvas).Control := self;

  ParentColor := False;
  sStyle := TsStyleDB.Create(Self);
  sStyle.COC := COC_TsDBComboBox;

  OnKeyDown := sStyle.OnKeyDown;
  FButtonWidth := 14;
  FDefListProc := nil;
  ItemHeight := 13;
  FColorArrow := clMenuText;
  if sStyle.Flat then begin
    FColorButton := sStyle.Painting.ColorInActive;
  end
  else begin
    FColorButton := clBtnFace;
  end;

  FSQL := TStringList.Create;
  CharsInCode := 0;
  UseItemIndex := False;
  DropDownCount := 16;
end;

destructor TsDBComboBox.Destroy;
begin
  SQL.Free;
  FreeObjectInstance(FListInstance);
  inherited Destroy;
end;

procedure TsDBComboBox.SetSQL(Value: TStrings);
begin
  SQL.Assign(Value);
end;

function TsDBComboBox.GetCurrentKod : integer;
begin
  Result := -1;
  if Active and (CharsInCode>0) and (Text <> '') then begin
    try
      Result := StrToInt(copy(Text, 1, CharsInCode));
    except
    end;
  end;
end;

function TsDBComboBox.IndexOfKod(s : string) : integer;
var
  i : integer;
  c1, c2 : string;
begin
  Result := -1;
  for i := 0 to Items.Count - 1 do begin
    c1 := AddChar('0', s, CharsInCode);
    c2 := ExtractWord(1, Items[i], [' ', '-']);
    if c1 = c2 then begin
      Result := i;
      Break;
    end;
  end;
end;

function TsDBComboBox.GetCurrentName : string;
begin
  if CharsInCode > 0 then begin
    Result := copy(Text, CharsInCode + 4, Length(Text) - CharsInCode - 3);
  end
  else begin
    Result := Text;
  end;
end;

procedure TsDBComboBox.SetActive(Value : boolean);
begin
  if Value then begin
    FActive := (Generate > 0);
  end
  else begin
    FActive := False;
  end;
end;

procedure TsDBComboBox.WMPaint(var Message: TWMPaint);
var
  R, bR: TRect;
  DC: HDC;
  PS: TPaintStruct;
  {BorderBrush, }WinBrush: HBRUSH;
begin
  if sStyle.Flat then begin
    DC := BeginPaint(Handle, PS);
    try
      R := PS.rcPaint;
      R.Right := iffi(R.Right > Width - FButtonWidth - 4, Width - FButtonWidth - 4, R.Right);
//      BorderBrush := CreateSolidBrush(ColorToRGB(sStyle.Painting.ColorBorderTop));
      WinBrush := CreateSolidBrush(ColorToRGB(sStyle.GetActiveColor));

      FillRect(DC, R, WinBrush);
      bR := GetButtonRect;
      if IntersectRect(bR, bR, PS.rcPaint) then PaintButton;
      ExcludeClipRect(DC, ClientWidth - FButtonWidth - 2, 0, ClientWidth, ClientHeight);
      PaintWindow(DC);
      if (Style = csDropDown) and DroppedDown then begin
        R := ClientRect;
        InflateRect(R, -2, -2);
        R.Right := Width - FButtonWidth - 8;
        Canvas.Brush.Color := sStyle.GetActiveColor;
        Canvas.FrameRect(R);
      end
      else if Style <> csDropDown then InvalidateSelection;
    finally
      EndPaint(Handle, PS);
    end;
    RedrawBorders;
    DeleteObject(WinBrush);
//    DeleteObject(BorderBrush);
    Message.Result := 0;
  end
  else inherited;
end;

//   
function TsDBComboBox.GetButtonRect: TRect;
begin
  GetWindowRect(Handle, Result);
  OffsetRect(Result, -Result.Left, -Result.Top);
  Inc(Result.Left, ClientWidth - FButtonWidth);
  OffsetRect(Result, -1, 0);
end;

//  
procedure TsDBComboBox.PaintButton;
var
  R: TRect;
  x, y, i: Integer;
  procedure PolyDown;
  begin
    canvas.Polygon([Point(x + 3, y + 3), Point(x + 9, y + 3), Point(x + 6, y)])
  end;
  procedure PolyUp;
  begin
    canvas.Polygon([Point(x + 3, y), Point(x + 9, y), Point(x + 6, y + 3)]);
  end;
begin
  R := GetButtonRect;
  InflateRect(R, 1, 0);

  Canvas.Brush.Color := FColorButton;
  Canvas.FillRect(R);
  Canvas.Brush.Color := sStyle.Painting.ColorBorderTop;
  i := 1;
  x := (R.Right - R.Left) div 2 - 6 + R.Left;
  y := (R.Bottom - R.Top) div 2 - 1 + R.Top;
  DrawRectangleOnDC(Canvas.Handle, R, sStyle.Painting.ColorBorderTop, sStyle.Painting.ColorBorderBottom, i);

  if Enabled then begin
    canvas.Brush.Color := FColorArrow;
    canvas.Pen.Color := FColorArrow;
    if DroppedDown then PolyDown else PolyUp;
  end
  else begin
    canvas.Brush.Color := clWhite;
    canvas.Pen.Color := clWhite;
    Inc(x); Inc(y);
    if DroppedDown then PolyDown else PolyUp;

//    PolyDown;
    Dec(x); Dec(y);
    canvas.Brush.Color := clGray;
    canvas.Pen.Color := clGray;
    if DroppedDown then PolyDown else PolyUp;
//    PolyDown;
  end;
  ExcludeClipRect(Canvas.Handle, ClientWidth - FButtonWidth, 0, ClientWidth, ClientHeight);
end;

procedure TsDBComboBox.CNCommand(var Message: TWMCommand);
var
  DC: HDC;
  WinBrush: HBRUSH;
  R: TRect;
begin
  inherited;
  if sStyle.Flat then begin
    Case Message.NotifyCode of
      CBN_SELCHANGE, CBN_SELENDOK, CBN_DROPDOWN: begin
        if not (Style in [csSimple, csDropDown]) then begin
          InvalidateSelection;
        end
        else if Message.NotifyCode = CBN_DROPDOWN then begin
          DC := GetWindowDC(Handle);
          WinBrush := CreateSolidBrush(ColorToRGB(sStyle.GetActiveColor));

          R := ClientRect;
          InflateRect(R, -2, -2);
          R.Top := R.Top + GetFontHeight(Font.Handle) - 5;
          FillRect(DC, R, WinBrush);

          R := ClientRect;
          R.Left := R.Left + 2 + self.Canvas.TextWidth(Text);
          FillRect(DC, R, WinBrush);

          ReleaseDC(Handle, DC);
          DeleteObject(WinBrush);
        end;
      end;
      CBN_CLOSEUP: begin
        R := GetButtonRect;
        R.Right := R.Left - 2;
        InvalidateRect(Handle, @R, FALSE);
      end;
    end;
  end;
end;

procedure TsDBComboBox.InvalidateSelection;
var
  DC: HDC;
  R: TRect;
  WinBrush: HBRUSH;
begin
  if sStyle.Flat then begin
    R := ClientRect;
    DC := GetWindowDC(Handle);
    R.Right := R.Right - FButtonWidth - 1;

    WinBrush := CreateSolidBrush(ColorToRGB(sStyle.Painting.ColorInActive));
    if (GetFocus = Handle) and not DroppedDown then begin
      InflateRect(R, -2, -2);
      FrameRect(DC, R, WinBrush);
      ReleaseDC(Handle, DC);
    end;
    DeleteObject(WinBrush);
  end;
end;

procedure TsDBComboBox.RedrawBorders;
begin
  Color := sStyle.GetActiveColor;
  sStyle.RedrawBorder;
  if (Style <> csSimple) then PaintButton;
end;

procedure TsDBComboBox.SetColors(Index: Integer; Value: TColor);
begin
  case Index of
    0: FColorArrow := Value;
    1: FColorButton := Value;
  end;
  Invalidate;
end;

procedure TsDBComboBox.WndProc(var Message: TMessage);
begin
  sStyle.WndProc(Message);
  case Message.Msg of
    WM_PARENTNOTIFY : begin
      case LoWord(Message.wParam) of
        WM_CREATE: begin
          if FDefListProc <> nil then begin
            SetWindowLong(FListHandle, GWL_WNDPROC, Longint(FDefListProc));
            FDefListProc := nil;
            FChildHandle := Message.lParam;
          end
          else
            if FChildHandle = 0 then
              FChildHandle := Message.lParam
            else
              FListHandle := Message.lParam;
        end
      end;
    end;
    WM_WINDOWPOSCHANGING : begin
      if sStyle.Flat and (Style in [csDropDown, csSimple]) then
        SetWindowPos( EditHandle, 0,
                      0, 0, ClientWidth - FButtonWidth - 2 * 2 - 4, Height - 2 * 2 - 2,
                      SWP_NOMOVE + SWP_NOZORDER + SWP_NOACTIVATE + SWP_NOREDRAW);
    end;
  end;
  inherited;
end;

procedure TsDBComboBox.Invalidate;
begin
  Color := sStyle.GetActiveColor;
  if sStyle.Flat then begin
    if Assigned(Parent) then RedrawBorders;
  end;
  inherited;
end;

procedure TsDBComboBox.MEnter(var Msg: TMessage);
begin
  sStyle.FFocused := True;
  invalidate;
  inherited;
end;

procedure TsDBComboBox.MExit(var Msg: TMessage);
begin
  sStyle.FFocused := False;
  invalidate;
  inherited;
end;

function TsDBComboBox.Generate: integer;
begin
  Result := 0;
end;
end.
